/*
 * abissd/sched.c - Scheduler module selection
 *
 * Written 2004 by Werner Almesberger
 *
 * (C) Copyright Koninklijke Philips Electronics NV 2004. All rights reserved.
 * For licensing and warranty information, see the file ../../COPYING.LGPLv21
 */


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>
#include <dlfcn.h>
#include <sys/types.h>

#include <abissd.h>

#include "common.h"


static struct sched {
    const struct sched_ops *ops;
    struct sched *next;
} *scheds = NULL;


void show_sched(FILE *file)
{
    const struct sched *walk;

    for (walk = scheds; walk; walk = walk->next)
	fprintf(file,"%c %s",walk == scheds ? ' ' : ',',walk->ops->name);
    fputc('\n',file);
}


const struct sched_ops *find_sched(const char *name)
{
    const struct sched *walk;

    for (walk = scheds; walk; walk = walk->next)
	if (!strcmp(walk->ops->name,name))
	    return walk->ops;
    return NULL;
}


static void register_sched(const struct sched_ops *ops)
{
    struct sched *sched;

    if (find_sched(ops->name)) {
	fprintf(stderr,"skipping duplicate module for \"%s\"\n",ops->name);
	return;
    }
    sched = malloc(sizeof(struct sched));
    if (!sched) {
	perror("malloc");
	exit(1);
    }
    sched->ops = ops;
    sched->next = scheds;
    scheds = sched;
    if (ops->init)
	ops->init();
}


void load_schedulers(const char *sched_path)
{
    char *path,*dir_name,*cwd;
    const char *license;

    cwd = getcwd(NULL,0); /* Linux extension */
    if (!cwd) {
	perror("getcwd");
	exit(1);
    }
    path = strdup(sched_path);
    if (!path) {
	perror("strdup");
	exit(1);
    }
    for (dir_name = strtok(path,":"); dir_name; dir_name = strtok(NULL,":")) {
	DIR *dir;

	dir = opendir(dir_name);
	if (!dir) {
	    perror(dir_name);
	    continue;
	}
	while (1) {
	    struct dirent *de;
	    void *handle;
	    int len;
	    const struct sched_ops *ops;

	    de = readdir(dir);
	    if (!de)
		break;
	    if (strncmp(de->d_name,"abissd_",7))
		continue;
	    len = strlen(de->d_name);
	    if (len < 4 || strcmp(de->d_name+len-3,".so"))
		continue;
	    if (*dir_name == '/') {
		path = malloc(strlen(dir_name)+1+strlen(de->d_name)+1);
		if (!path) {
		    perror("malloc");
		    exit(1);
		}
		sprintf(path,"%s/%s",dir_name,de->d_name);
	    }
	    else {
		path = malloc(strlen(cwd)+1+strlen(dir_name)+1+
		  strlen(de->d_name)+1);
		if (!path) {
		    perror("malloc");
		    exit(1);
		}
		sprintf(path,"%s/%s/%s",cwd,dir_name,de->d_name);
	    }
	    handle = dlopen(path,RTLD_NOW);
	    if (!handle) {
		fprintf(stderr,"%s: %s\n",path,dlerror());
		continue;
	    }
	    ops = dlsym(handle,"sched_ops");
	    if (!ops) {
		fprintf(stderr,"%s: %s\n",path,dlerror());
		(void) dlclose(handle);
		continue;
	    }
	    license = dlsym(handle,"license");
	    if (!license) {
		fprintf(stderr,"%s: %s\n",path,dlerror());
		(void) dlclose(handle);
		continue;
	    }
	    if (verbose)
		fprintf(stderr,"registering \"%s\": %s (%s)\n",ops->name,path,
		  license);
	    register_sched(ops);
	    /* don't dlclose ! */
	}
	if (closedir(dir) < 0) {
	    perror("closedir");
	    exit(1);
	}
    }
    free(cwd);
    free(path);
    if (!scheds)
	fprintf(stderr,"warning: no scheduler plugins found in %s\n",
	  sched_path);
}
