/*
 * (C) Copyright Koninklijke Philips Electronics NV 2004, 2005. All rights
 * reserved. For licensing and warranty information, see the file
 * ../../COPYING.LGPLv21
 */

/*
 * History:	05/03/04 - Chris Ross - Initial Version
 *		11/08/04 - Erik Bos - Added file descripter interface
 */

#include "libabiss.h"

#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <abiss_msg.h>
#include <abiss_msg_basic.h>

const int DEBUG = 1;
const char *DEFAULT_SCHEDULER = "basic";

static int attach( int fd, const int byterate, const int app_buffer )
{
	struct abiss_attach_msg		 realtime;
	struct abiss_sched_basic_prm	 basic_options;
	struct abiss_set_msg		 query = { .header = 
	                                         { .type = abiss_query }};

	/* Find out which scheduler abiss is using */
	if ( ioctl(fd, ABISS_IOCTL, &query ) < 0 ) {
		/* Not an abiss file. */
		perror("ABISS_IOCTL (abiss_query)");
		return 0;
	}

	if ( DEBUG ) {
		printf("attach: scheduler: '%s'\n",
		       *query.sched_name ? query.sched_name : "(none)");
	}

	/* Clear the abiss structures the brute force way... */
	memset( &realtime, 0, sizeof( struct abiss_attach_msg ));
	memset( &basic_options, 0, sizeof( struct abiss_sched_basic_prm ));

	/* Choose our scheduler */
	strncpy( realtime.sched_name,
		 (*query.sched_name ? query.sched_name : DEFAULT_SCHEDULER),
		 ABISS_NAME_LENGTH );

	/* Call abiss to associate the file with the playout buffer
	 *
	 * Note that the application has alreay included any deviations from
	 * its own side in "app_buffer" (and maybe also in "byterate"), e.g.
	 * using a formula like this
	 *
	 *	 app_buffer = read_buffer_size+app_jitter*rate+
	 *                    app_non_sequentiality;
	 */

	basic_options.ra_bytes = app_buffer;
	basic_options.fill_Bps = byterate;
	realtime.sched_prm = &basic_options;	

	realtime.header.type = abiss_attach;
	

	if ( 0 > ioctl( fd, ABISS_IOCTL, &realtime ) ) {
		return -1;
	}

	return 0;
}

static int updateplayoutpoint_fd( int fd, int mode )
{
	struct abiss_position_msg position = {
		.header.type	= abiss_position,
		.whence		= SEEK_CUR,
		.seek_mode	= mode, 
		.pos		= 0
	};

	if ( ioctl( fd, ABISS_IOCTL, &position ) ) {
		fprintf( stderr, "updateplayoutpoint_fd: "
			"Couldn't update playout point\n" );
		return -1;
	}
	return 0;
}

static int updateplayoutpoint_fp( FILE *fp, int mode )
{
	return updateplayoutpoint_fd( fileno(fp), mode );
}


/* ------------- exported functions ------------ */

off_t abiss_get_playoutpoint_fd( int fd )
{
	struct abiss_set_msg query;

	/* Find out which scheduler abiss is using */
	query.header.type = abiss_query;
	if ( ioctl(fd, ABISS_IOCTL, &query ) < 0 ) {
		perror("ABISS_IOCTL (abiss_query)");
		return -1;
	}

	if ( !strcmp( query.sched_name, "basic" ) ) {
		struct abiss_sched_basic_msg basic_query;

		basic_query.header.type = abiss_sched_msg;
		basic_query.type = abiss_sb_query;
		basic_query.u.query.page_state = NULL;
		basic_query.u.query.pages = 0;
		
		if ( ioctl(fd, ABISS_IOCTL, &basic_query ) < 0 ) {
			perror("ABISS_IOCTL (abiss_query)");
			return -1;
		}
		return (off_t)basic_query.u.query.playout;
	}
	abort();
}

off_t abiss_get_playoutpoint_fp( FILE *fp )
{
	return abiss_get_playoutpoint_fd( fileno(fp) );
}

FILE* abiss_fopen( const char *path, const char *mode, 
		   const int byterate, const int app_buffer )
{
	FILE *fp;
	int fd;

	/* First up, do a normal fopen. */
	fp = fopen( path, mode );

	if ( fp == NULL )
		return NULL;
	
	fd = fileno(fp);
	if ( fd == -1 ) {
		fprintf( stderr, "abiss_fopen: couldn't convert to fd\n");
		fclose( fp );
		return NULL;
	}
	else
		fprintf( stderr, "abiss_fopen: fileno(fp) = %d\n", fd);

	/*
	 * For buffered IO, we also need to take into account additional data
	 * read due to buffering. POSIX 1003.1 says that "BUFSIZ" is the "Size
	 * of <stdio.h> buffers.". In POSIX we trust, so be it :-)
	 *
	 * Note that applications later calling setbuf, setvbuf, etc. need to
	 * consider the buffer size change. (Better yet, applications should
	 * probably just not try to change the buffer size.)
	 */

	if ( attach(fd, byterate, app_buffer+BUFSIZ-1 ) == -1 ) {
		fprintf( stderr, 
			 "abiss_fopen: couldn't get requested rate/buffer of "
			 "%d/%d for %s\n", byterate, app_buffer, path );
		fclose( fp );
		fp = NULL;
	}

	fprintf( stderr, "abiss_fopen: fp = '%p'\n", fp );

	return fp;
}

int abiss_fclose( FILE *fp )
{
	return fclose( fp );
}

size_t abiss_fread( void *ptr, size_t size, size_t nmemb, FILE *fp )
{
	size_t read = fread( ptr, size, nmemb, fp );

	if ( read > 0 ) 
		updateplayoutpoint_fp( fp, abiss_seek_rt ); 

	return read;
}

size_t abiss_fwrite( const void *ptr, size_t size, size_t nmemb, FILE *fp )
{
	return fwrite( ptr, size, nmemb, fp );
}

int abiss_fseek( FILE *fp, long offset, int whence )
{
	int result = fseek( fp, offset, whence );

	if ( 0 == result )
		result = updateplayoutpoint_fp( fp, abiss_seek_be);

	return result;
}


int abiss_fseeko( FILE *fp, off_t offset, int whence )
{
	int result = fseeko( fp, offset, whence );

	if ( 0 == result )
		result = updateplayoutpoint_fp( fp, abiss_seek_be );
	
	return result;
}

long abiss_ftell( FILE *fp )
{
	/* Nothing special to do for tell */
	return ftell( fp );
}


off_t abiss_ftello( FILE *fp )
{
	/* Nothing special to do for tello */
	return ftello( fp );
}

int abiss_open( const char *path, int flags, mode_t mode,
		const int byterate, const int app_buffer )
{
	int fd;

	/* First up, do a normal fopen. */
	fd = open(path, flags, mode);
	if ( fd == -1 )
		return -1;
	
	if ( attach(fd, byterate, app_buffer) == -1 ) {
		fprintf( stderr, 
			 "abiss_open: couldn't get requested rate/buffer of "
			 "%d/%d for %s\n", byterate, app_buffer, path );
		close(fd);
		fd = -1;
	}
	return fd;
}

ssize_t abiss_read( int fd, void *buf, size_t count)
{
	ssize_t bytes_read = read( fd, buf, count );

	if ( bytes_read > 0 ) 
		updateplayoutpoint_fd( fd, abiss_seek_rt );
 
	return bytes_read;
}

ssize_t abiss_write( int fd, void *buf, size_t count)
{
	return write( fd, buf, count );
}

off_t abiss_lseek( int fd, off_t offset, int whence )
{
	off_t result = lseek( fd, offset, whence );

	if ( result >= 0 )
		if ( 0 != updateplayoutpoint_fd( fd, abiss_seek_be ) )
			return -1;
				
	return result;
}

int abiss_close( int fd )
{
	return close( fd );
}
