/*
 * racc_posix.c - Random reader/writer, POSIX AIO interface
 *
 * Written 2004 by Werner Almesberger
 *
 * (C) Copyright Koninklijke Philips Electronics NV 2004. All rights reserved.
 * For licensing and warranty information, see the file ../../COPYING.GPLv2
 */

/*
 * aio_init is a GNU extension and isn't normally documented in man pages.
 * Here's the description:
 * http://www.gnu.org/software/libc/manual/html_mono/libc.html.gz#Asynchronous%20I%2FO
 */


#define _XOPEN_SOURCE 500 /* or else we don't get AIO */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <aio.h>
#include <sys/types.h>
#include <sys/mman.h>

#include "racc.h"


#define MAX_AIO_THREADS 1000


static void do_io(struct aiocb **list,int nr)
{
    int i;

    if (lio_listio(LIO_WAIT,list,nr,NULL) < 0) {
	perror("lio_listio");
	exit(1);
    }
    for (i = 0; i != nr; i++)
	if (aio_error(list[i])) {
	    fprintf(stderr,"request %d: %s\n",i,strerror(aio_error(list[i])));
	    exit(1);
	}
}


void racc_posix(int fd,int flags,off_t obj_size,int access_size,
  int step_size,int accesses,int locations)
{
    static struct aioinit aioinit;
    struct aiocb *aiocb,**list;
    void *buf;
    int i;

    aioinit.aio_threads =
      accesses < MAX_AIO_THREADS ? accesses : MAX_AIO_THREADS;
    aioinit.aio_num = accesses;
    aio_init(&aioinit);

    buf = calloc((flags == O_RDWR ? accesses : 1)*access_size+PAGE-1,1);
    if (!buf) {
	perror("calloc");
	exit(1);
    }
    buf = (void *) ((unsigned long) buf & ~(PAGE-1));
    aiocb = calloc(accesses,sizeof(struct aiocb));
    if (!aiocb) {
	perror("calloc");
	exit(1);
    }
    list = malloc(sizeof(struct aiocb *)*accesses);
    if (!list) {
	perror("malloc");
	exit(1);
    }
    for (i = 0; i != accesses; i++) {
	list[i] = aiocb+i;
	aiocb[i].aio_fildes = fd;
	aiocb[i].aio_buf = flags == O_RDWR ? buf+access_size*i : buf;
	aiocb[i].aio_nbytes = access_size;
	aiocb[i].aio_offset = (off_t) step_size*(rand() % locations);
    }
    if (mlockall(MCL_CURRENT) < 0) {
	perror("mlockall");
	exit(1);   
    }           
    if (flags == O_RDONLY || flags == O_RDWR) {
	for (i = 0; i != accesses; i++)
	    aiocb[i].aio_lio_opcode = LIO_READ;
	if (verbose)
	    show_time("read setup");
	do_io(list,accesses);
	if (verbose)
	    show_time("read");
    }
    if (flags == O_WRONLY || flags == O_RDWR) {
	for (i = 0; i != accesses; i++)
	    aiocb[i].aio_lio_opcode = LIO_WRITE;
	if (verbose)
	    show_time("write setup");
	do_io(list,accesses);
	if (verbose)
	    show_time("write");
    }
}
