/*
 * watchf.c - "watch" a file
 *
 * Written 2004 by Werner Almesberger
 *
 * (C) Copyright Koninklijke Philips Electronics NV 2004. All rights reserved.
 * For licensing and warranty information, see the file ../../COPYING.GPLv2
 */


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>


#define BUF 8192


static void do_gettimeofday(struct timeval *tv)
{
    if (gettimeofday(tv,NULL) < 0) {
	perror("gettimeofday");
	exit(1);
    }
}


static void usage(const char *name)
{
    fprintf(stderr,
"usage: %s [-m] [-n ...] [-t] file [seconds]\n\n"
"  -m  lock all pages in physical memory, so that reports are not delayed\n"
"      by high memory pressure (requires privileges)\n"
"  -n  output one new newline after each copy of the file content. Repeating\n"
"      this option increases the number of newlines.\n"
"  -t  print the current time before each copy of the file content\n"
  ,name);
    exit(1);
}


int main(int argc,char **argv)
{
    struct timeval interval = {
	.tv_sec = 1,
	.tv_usec = 0,
    };
    const char *name;
    double f;
    int fd;
    char buf[BUF];
    struct timeval start,due;
    int show_time = 0,do_mlock = 0,nl = 0;

    name = *argv;
    while (argv[1] && *argv[1] == '-') {
	char *c = argv[1]+1;

	do switch (*c) {
		case 't':
		    show_time = 1;
		    break;
		case 'm':
		    do_mlock = 1;
		    break;
		case 'n':
		    nl++;
		    break;
		default:
		    usage(name);
	    }
	while (*++c);
	argv++;
	argc--;
    }
    switch (argc) {
	char *end;

	case 3:
	    f = strtod(argv[2],&end);
	    if (*end)
		usage(name);
	    interval.tv_sec = f;
	    interval.tv_usec = (f-interval.tv_sec)*1000000;
	    /* fall through */
	case 2:
	    fd = open(argv[1],O_RDONLY);
	    if (fd < 0) {
		perror(argv[1]);
		return 1;
	    }
	    break;
	default:
	    usage(name);
    }

    /*
     * This isn't entirely "clean", since we have plenty of data that we may
     * not have touched yet, but it usually yields satisfactory results.
     */
    if (do_mlock)
	if (mlockall(MCL_CURRENT | MCL_FUTURE) < 0) {
	    perror("mlockall");
	    return 1;
	}

    do_gettimeofday(&start);
    due = start;
    while (1) {
	struct timeval now,nap;
	int nl_left;

	if (show_time) {
	    do_gettimeofday(&now);
	    now.tv_sec -= start.tv_sec;
	    now.tv_usec -= start.tv_usec;
	    if (now.tv_usec < 0) {
		now.tv_sec--;
		now.tv_usec += 1000000;
	    }
	    printf("%ld.%06ld: ",now.tv_sec,now.tv_usec);
	    fflush(stdout);
	}

	while (1) {
	    ssize_t got;

	    got = read(fd,buf,BUF);
	    if (got < 0) {
		perror("read");
		return 1;
	    }
	    if (!got)
		break;
	    while (got) {
		ssize_t wrote;

		wrote = write(1,buf,got);
		if (wrote < 0) {
		    perror("write");
		    return 1;
		}
		if (!wrote) {
		    fprintf(stderr,"EOF on write\n");
		    return 1;
		}
		got -= wrote;
	    }
	}
	if (lseek(fd,0,SEEK_SET) == (off_t) -1) {
	    perror("lseek");
	    return 1;
	}

	nl_left = nl;
	while (nl_left) {
	    ssize_t wrote;

	    wrote = write(1,"\n\n\n\n",nl_left > 4 ? 4 : nl_left);
	    if (wrote < 0) {
		perror("write");
		return 1;
	    }
	    if (!wrote) {
		fprintf(stderr,"EOF on write\n");
		return 1;
	    }
	    nl_left -= wrote;
	}

	due.tv_sec += interval.tv_sec;
	due.tv_usec += interval.tv_usec;
	if (due.tv_usec > 999999) {
	    due.tv_sec++;
	    due.tv_usec -= 1000000;
	}
	nap = due;
	do_gettimeofday(&now);
	nap.tv_sec -= now.tv_sec;
	nap.tv_usec -= now.tv_usec;
	if (nap.tv_usec < 0) {
	    nap.tv_sec--;
	    nap.tv_usec += 1000000;
	}
	if (nap.tv_sec < 0) {
	    if (nap.tv_usec) {
		nap.tv_usec = 1000000-nap.tv_usec;
		nap.tv_sec++;
	    }
	    fprintf(stderr,"Skipped %ld.%06ld s\n",-nap.tv_sec,nap.tv_usec);
	    due = now;
	}
	else {
	    struct timespec req;

	    req.tv_sec = nap.tv_sec;
	    req.tv_nsec = nap.tv_usec*1000;
	    if (nanosleep(&req,NULL) < 0) {
		perror("nanosleep");
		return 1;
	    }
	}
    }
    return 0; /* not reached */
}
