/*
 * tests/elv/Kglue.c - Interface with the "kernel"
 *
 * Written 2004-2005 by Werner Almesberger
 *
 * (C) Copyright Koninklijke Philips Electronics NV 2004-2005. All rights
 * reserved. For licensing and warranty information, see the file
 * ../../COPYING.GPLv2
 *
 * Some items copied from kernel files have other/additional copyright holders.
 */


#include <stdarg.h>
#include <assert.h>

#include "Kglue.h"


/* ----- Block device callbacks -------------------------------------------- */


static int prep_rq_fn(struct request_queue *q,struct request *rq)
{
	return BLKPREP_OK;
}


/*
 * The following functions (from ll_rw_blk.c) just check that we're not
 * exceeding various hardware limits, and update hardware-related counters.
 * None of this is of any interest in our context.
 */


static int merge_requests_fn(request_queue_t *q, struct request *req,
    struct request *next)
{
	return 1;
}


/* ----- Other helper functions -------------------------------------------- */


/* __blk_put_request, blk_free_request */


static void __blk_put_request(request_queue_t *q, struct request *rq)
{
	struct request_list *rl = rq->rl;

	if (!q)
		return;
	rq->q = NULL;
	rq->rl = NULL;
	if (rl) {
		elv_completed_request(q,rq);
		BUG_ON(!list_empty(&rq->queuelist));
		elv_put_request(q,rq);
	}
	kfree(rq);
}


static void free_bio(struct request *rq)
{
	struct bio *bio;

	bio = rq->bio;
	while (bio) {
		struct bio *next = bio->bi_next;

		kfree(bio->bi_io_vec);
		kfree(bio);
		if (bio == rq->biotail)
			break;
		bio = next;
	}
}


void blkdev_dequeue_request(struct request *rq)
{
        BUG_ON(list_empty(&rq->queuelist));
        list_del_init(&rq->queuelist);
        if (rq->rl)
		elv_remove_request(rq->q,rq);
}


/* ----- Our kernel's data structures -------------------------------------- */


static struct request_list rl;


static struct request_queue q = {
	.prep_rq_fn = prep_rq_fn,
	.merge_requests_fn = merge_requests_fn,
	.ordered = QUEUE_ORDERED_TAG,
};


static struct gendisk gendisk;


static struct block_device bdev = {
	.q = &q,
	.bd_disk = &gendisk,
};


static struct inode inode_normal = {
	.i_abiss = 0,
};


static struct inode inode_abiss = {
	.i_abiss = 1,
};


static struct address_space mapping_normal = {
	.host = &inode_normal,
};


static struct address_space mapping_abiss = {
	.host = &inode_abiss,
};


struct super_block super = {
	.s_bdev = &bdev,
};


#define PAGES 10000
#define MAX_SG 1000

static struct page pages[PAGES];


void init_kernel(void)
{
	extern void __init_fn(void);
	extern void __setup_fn__elevator(char *s);

	/* don't want "io scheduler abiss registered" */
	no_printk = 1;
	__init_fn();
	no_printk = 0;
	__setup_fn__elevator("abiss");
	elevator_init(&q,"abiss");
	prio_tree_init();
}


void shutdown_kernel(void)
{
	elevator_exit(q.elevator);
}


/* ----- Things from ll_rw_blk.c ------------------------------------------- */


#include "_ll_rw_blk.c"


/* See ll_rw_blk.c:__make_request */


static struct request *make_request(int flags,request_queue_t *q,
    struct bio *bio)
{
	sector_t sector = 0;
	int nr_sectors = 0;
	struct request *req;
	int barrier,el_ret;

	if (bio) {
		sector = bio->bi_sector;
		nr_sectors = bio->bi_size >> 9;
	}
	barrier = flags & (REQ_SOFTBARRIER | REQ_HARDBARRIER);
	if (barrier || !bio)
		goto get_rq;

	el_ret = elv_merge(q, &req, bio);
	switch (el_ret) {
		case ELEVATOR_BACK_MERGE:
			BUG_ON(!rq_mergeable(req));
			req->biotail->bi_next = bio;
                        req->biotail = bio;
                        req->nr_sectors += nr_sectors;
			if (!attempt_back_merge(q, req))
				elv_merged_request(q, req);
			return NULL;
			
		case ELEVATOR_FRONT_MERGE:
			BUG_ON(!rq_mergeable(req));
			bio->bi_next = req->bio;
			req->bio = bio;
			req->sector = sector;
			req->nr_sectors += nr_sectors;
			if (!attempt_front_merge(q, req))
				elv_merged_request(q, req);
			return NULL;

		case ELEVATOR_NO_MERGE:
			break;

		default:
			BUG();
	}

get_rq:
	req = kmalloc(sizeof(struct request),GFP_KERNEL);
	req->flags = flags;
	if (barrier)
		req->flags |= REQ_NOMERGE;
	INIT_LIST_HEAD(&req->queuelist);
	req->elevator_private = NULL;
	req->sector = sector;
	req->nr_sectors = nr_sectors;
	req->bio = req->biotail = bio;
	req->q = q;
	req->rl = flags & (REQ_PM_SUSPEND | REQ_PM_RESUME | REQ_DRIVE_CMD) ?
	    NULL : &rl;
	req->rq_disk = &gendisk;
	req->waiting = req->special = req->data_len = req->start_time = 0;
	if (req->rl)
		elv_set_request(q,req,GFP_KERNEL);

	return req;
}


/* ----- Queue access ------------------------------------------------------ */


static void request_done(struct request *rq)
{
	blkdev_dequeue_request(rq);
	free_bio(rq);
	__blk_put_request(&q,rq);
}


struct request *concoct_rq(int flags,int start,int secs)
{
	struct bio *bio = NULL;
	struct page *page = NULL;
	int i;

	if (!secs)
		goto just_rq;

	bio = kmalloc(sizeof(struct bio),GFP_KERNEL);
	bio->bi_rw = flags;
	bio->bi_vcnt = 0;
	bio->bi_io_vec = kmalloc(sizeof(struct bio_vec)*MAX_SG,GFP_KERNEL);
	bio->bi_next = NULL;
	bio->bi_sector = 0;
	bio->bi_size = 0;
	bio->bi_bdev = &bdev;
	bio_set_prio(bio,ABISS_PRIO_MAX-
	    (flags & REQ_ABISS ? REQ_ABISS_GET_PRIO(flags) : ABISS_PRIO_MAX));

	for (i = start; i != start+secs; i++) {
		if (page != pages+(i >> 3)) {
			if (page)
				bio->bi_vcnt++;
			page = pages+(i >> 3);
			page->mapping = flags & REQ_ABISS ?
			    &mapping_abiss : &mapping_normal;
			bio->bi_io_vec[bio->bi_vcnt].bv_page = page;
			bio->bi_io_vec[bio->bi_vcnt].bv_offset =
			    (i & 7) << 9;
			bio->bi_io_vec[bio->bi_vcnt].bv_len = 0;
		}
		if (!bio->bi_size)
			bio->bi_sector = start;
		bio->bi_io_vec[bio->bi_vcnt].bv_len += 512;
		bio->bi_size += 512;
	}
	bio->bi_vcnt++;

just_rq:
	return make_request(flags,&q,bio);
}


void enqueue(struct request *rq)
{
	if (verbose)
		printk("ENQUEUE %p (%lu+%lu) flags 0x%x\n",rq,
		    (unsigned long) rq->sector,(unsigned long) rq->nr_sectors,
		    rq->flags);

	elv_add_request(&q,rq,
	    rq->flags & REQ_ABISS_HEAD ? ELEVATOR_INSERT_FRONT :
	    rq->flags & REQ_ABISS_TAIL ? ELEVATOR_INSERT_BACK :
	    ELEVATOR_INSERT_SORT,0);
}


void dequeue(int mode,struct request *rq,const char *id,const char *file,
    int line,...)
{
	struct request *got;
	va_list ap;
	int start,secs,check;

	if (dump_queue)
		return;

	BUG_ON(elv_queue_empty(&q));

	va_start(ap,line);
	start = va_arg(ap,int);
	secs = va_arg(ap,int);
	check = va_arg(ap,int);
	va_end(ap);

	if (start == -1 || secs == -1 || check != -1) {
		if (start != -1 || secs != -1 || check != -1)
			panic("%s:%d: wrong number of arguments\n",file,line);
		start = secs = 0;
	}

	got = mode & RQ_GET ? elv_next_request(&q) : rq;
	if (!got)
		panic("%s:%d: expected \"%s\" (%d+%d), got NULL\n",
		    file,line,id,start,secs);

	if (verbose)
		printk("%s %p (%lu+%lu)\n",
		    mode == (RQ_GET | RQ_END) ? "DEQUEUE" :
		    mode == RQ_GET ? "GET" : "END",rq,
		    (unsigned long) rq->sector,
		    (unsigned long) rq->nr_sectors);

	if (rq != got)
		panic("%s:%d: expected \"%s\" (%p, %d+%d), got %p (%lu+%lu)\n",
		    file,line,id,rq,start,secs,got,
		    (unsigned long) got->sector,
		    (unsigned long) got->nr_sectors);

	if (secs && (rq->sector != start || rq->nr_sectors != secs))
		panic("%s:%d: expected \"%s\" to be %d+%d, is %lu+%lu\n",
		    file,line,id,start,secs,
		    (unsigned long) rq->sector,(unsigned long) rq->nr_sectors);

	if (mode & RQ_END)
		request_done(got);
}


void upgrade(int offset,int prio)
{
	/* offset is in sectors ! */

	if (verbose)
		printk("UPGRADE %d prio %lu\n",offset,prio);
	elevator_upgrade(&super,offset,8,prio);
}


void queue_empty(void)
{
	assert(!elv_next_request(&q));
}


void phase(int rw)
{
	extern void set_phase(struct request_queue *q,int rw);

	set_phase(&q,rw);
}


void show_queue(void)
{
	struct request *rq;

	while (1) {
		rq = elv_next_request(&q);
		if (!rq)
			break;
		printk("DEQUEUE %p (%lu+%lu) flags 0x%x\n",rq,
		    (unsigned long) rq->sector,(unsigned long) rq->nr_sectors,
		    rq->flags);
		request_done(rq);
	}
}
